@@ -470,10 +470,11 @@ Hexdump.hexdump( System.err, src, srcIndex, maxLen < 128 ? maxLen + 8 : 128 );
             default:
                 c = "UNKNOWN";
         }
+        String str = errorCode == 0 ? "0" : SmbException.getMessageByCode( errorCode );
         return new String(
             "command="      + c +
             ",received="    + received +
-            ",errorCode=" + SmbException.getMessageByCode( errorCode ) +
+            ",errorCode="   + str +
             ",flags=0x"     + Hexdump.toHexString( flags & 0xFF, 4 ) +
             ",flags2=0x"    + Hexdump.toHexString( flags2, 4 ) +
             ",signSeq="     + signSeq +
