@@ -1373,7 +1373,11 @@ public class SmbFile extends URLConnection {
         if( dfsReferral == null ) {
             return null;
         }
-        return "smb:/" + (new String( dfsReferral.node + unc )).replace( '\\', '/' );
+        String path = "smb:/" + (new String( dfsReferral.node + unc )).replace( '\\', '/' );
+        if (isDirectory()) {
+            path += '/';
+        }
+        return path;
     }
 
 /**
