@@ -128,7 +128,7 @@ public class SmbTransport extends Transport implements SmbConstants {
             }
         }
 
-                                        /* close old sessions */
+                                        /* logoff old sessions */
         if (SO_TIMEOUT > 0 && sessionExpiration < (now = System.currentTimeMillis())) {
             sessionExpiration = now + SO_TIMEOUT;
             iter = sessions.listIterator();
@@ -136,7 +136,6 @@ public class SmbTransport extends Transport implements SmbConstants {
                 ssn = (SmbSession)iter.next();
                 if( ssn.expiration < now ) {
                     ssn.logoff( false );
-                    iter.remove();
                 }
             }
         }
@@ -148,10 +147,10 @@ public class SmbTransport extends Transport implements SmbConstants {
         return ssn;
     }
     boolean matches( UniAddress address, int port, InetAddress localAddr, int localPort ) {
-        int p1 = ( port == 0 || port == DEFAULT_PORT ) ? 0 : port;
-        int p2 = ( this.port == 0 || this.port == DEFAULT_PORT ) ? 0 : this.port;
         return address.equals( this.address ) &&
-                    p1 == p2 &&
+                    (port == 0 || port == this.port ||
+                            /* port 139 is ok if 445 was requested */
+                            (port == 445 && this.port == 139)) &&
                     (localAddr == this.localAddr ||
                             (localAddr != null &&
                                     localAddr.equals( this.localAddr ))) &&
@@ -226,9 +225,18 @@ public class SmbTransport extends Transport implements SmbConstants {
          * until we have properly negotiated.
          */
         synchronized (sbuf) {
+            /* If jcifs.netbios.hostname is set this *probably* means there
+             * is a policy regarding which hosts a user can connect from. This
+             * requires communicating over port 139 rather than 445.
+             */
+            if (NETBIOS_HOSTNAME != null && NETBIOS_HOSTNAME.equals( "" ) == false) {
+                port = 139;
+            }
             if (port == 139) {
                 ssn139();
             } else {
+                if (port == 0)
+                    port = DEFAULT_PORT; // 445
                 if (localAddr == null) {
                     socket = new Socket( address.getHostAddress(), port );
                 } else {
@@ -243,18 +251,34 @@ public class SmbTransport extends Transport implements SmbConstants {
             NEGOTIATE_REQUEST.mid = mid;
             int n = NEGOTIATE_REQUEST.encode( sbuf, 4 );
             Encdec.enc_uint32be( n & 0xFFFF, sbuf, 0 ); /* 4 byte ssn msg header */
+
+            if (log.level > 3) {
+                log.println( NEGOTIATE_REQUEST );
+                if (log.level > 5) {
+                    Hexdump.hexdump( log, sbuf, 4, n );
+                }
+            }
+
             out.write( sbuf, 0, 4 + n );
             out.flush();
             /* Note the Transport thread isn't running yet so we can
              * read from the socket here.
              */
-            peekKey(); /* try to read header */
-            int size = Encdec.dec_uint16be( sbuf, 2 );
+            if (peekKey() == null) /* try to read header */
+                throw new IOException( "transport closed in negotiate" );
+            int size = Encdec.dec_uint16be( sbuf, 2 ) & 0xFFFF;
             if (size < 33 || (4 + size) > sbuf.length ) {
                 throw new IOException( "Invalid payload size: " + size );
             }
             readn( in, sbuf, 4 + 32, size - 32 );
             resp.decode( sbuf, 4 );
+
+            if (log.level > 3) {
+                log.println( resp );
+                if (log.level > 5) {
+                    Hexdump.hexdump( log, sbuf, 4, n );
+                }
+            }
         }
     }
     public void connect() throws SmbException {
@@ -273,7 +297,7 @@ public class SmbTransport extends Transport implements SmbConstants {
         try {
             negotiate( port, resp );
         } catch( ConnectException ce ) {
-            port = port == 445 ? 139 : 445;
+            port = (port == 0 || port == DEFAULT_PORT) ? 139 : DEFAULT_PORT;
             negotiate( port, resp );
         }
 
@@ -308,7 +332,6 @@ public class SmbTransport extends Transport implements SmbConstants {
         while (iter.hasNext()) {
             SmbSession ssn = (SmbSession)iter.next();
             ssn.logoff( hard );
-            iter.remove();
         }
         socket.shutdownOutput();
         out.close();
@@ -325,12 +348,12 @@ public class SmbTransport extends Transport implements SmbConstants {
     protected Request peekKey() throws IOException {
         int n;
         do {
-            n = readn( in, sbuf, 0, 4 );
-        } while (sbuf[0] == 0x85);  /* Dodge NetBIOS keep-alive */
-                                            /* read smb header */
-        if ((n = readn( in, sbuf, 4, 32 )) < 32) {
-            return null;                      /* stream closed */
-        }
+            if ((n = readn( in, sbuf, 0, 4 )) < 4)
+                return null;
+        } while (sbuf[0] == (byte)0x85);  /* Dodge NetBIOS keep-alive */
+                                                   /* read smb header */
+        if ((n = readn( in, sbuf, 4, 32 )) < 32)
+            return null;
         if (log.level > 2) {
             log.println( "New data read: " + this );
             jcifs.util.Hexdump.hexdump( log, sbuf, 4, 32 );
@@ -354,15 +377,15 @@ public class SmbTransport extends Transport implements SmbConstants {
             }
                                         /* out of phase maybe? */
                           /* inch forward 1 byte and try again */
-            for (int i = 0; i < 7; i++) {
+            for (int i = 0; i < 35; i++) {
                 sbuf[i] = sbuf[i + 1];
             }
             int b;
-            if ((b = in.read()) == 0) return null;
-            sbuf[7] = (byte)b;
+            if ((b = in.read()) == -1) return null;
+            sbuf[35] = (byte)b;
         }
 
-        key.mid = Encdec.dec_uint16le( sbuf, 34 );
+        key.mid = Encdec.dec_uint16le( sbuf, 34 ) & 0xFFFF;
 
         /* Unless key returned is null or invalid Transport.loop() always
          * calls doRecv() after and no one else but the transport thread
@@ -390,6 +413,20 @@ public class SmbTransport extends Transport implements SmbConstants {
             out.write( BUF, 0, 4 + n );
         }
     }
+    protected void doSend0( Request request ) throws IOException {
+        try {
+            doSend( request );
+        } catch( IOException ioe ) {
+            if (log.level > 2)
+                ioe.printStackTrace( log );
+            try {
+                disconnect( true );
+            } catch( IOException ioe2 ) {
+                ioe2.printStackTrace( log );
+            }
+            throw ioe;
+        }
+    }
 
     protected void doRecv( Response response ) throws IOException {
         ServerMessageBlock resp = (ServerMessageBlock)response;
@@ -397,7 +434,7 @@ public class SmbTransport extends Transport implements SmbConstants {
 
         synchronized (BUF) {
             System.arraycopy( sbuf, 0, BUF, 0, 4 + HEADER_LENGTH );
-            int size = Encdec.dec_uint16be( BUF, 2 );
+            int size = Encdec.dec_uint16be( BUF, 2 ) & 0xFFFF;
             if (size < (HEADER_LENGTH + 1) || (4 + size) > rcv_buf_size ) {
                 throw new IOException( "Invalid payload size: " + size );
             }
@@ -419,13 +456,17 @@ public class SmbTransport extends Transport implements SmbConstants {
                 }
             }
 
-            if (digest != null) {
+            /* Verification fails (w/ W2K3 server at least) if status is not 0. This
+             * suggests MS doesn't compute the signature (correctly) for error responses
+             * (perhaps for DOS reasons).
+             */
+            if (digest != null && resp.errorCode == 0) {
                 digest.verify( BUF, 4, resp );
             }
         }
     }
     protected void doSkip() throws IOException {
-        int size = Encdec.dec_uint16be( sbuf, 2 );
+        int size = Encdec.dec_uint16be( sbuf, 2 ) & 0xFFFF;
         if (size < 33 || (4 + size) > rcv_buf_size ) {
             /* log message? */
             in.skip( in.available() );
@@ -477,7 +518,7 @@ public class SmbTransport extends Transport implements SmbConstants {
 
         try {
             if (response == null) {
-                doSend( request );
+                doSend0( request );
                 return;
             } else if (request instanceof SmbComTransaction) {
                 response.command = request.command;
@@ -517,7 +558,7 @@ public class SmbTransport extends Transport implements SmbConstants {
                              */
 
                             do {
-                                doSend( req );
+                                doSend0( req );
                             } while( req.hasMoreElements() && req.nextElement() != null );
 
                             /* 
@@ -567,7 +608,7 @@ public class SmbTransport extends Transport implements SmbConstants {
 
     /* DFS */
 
-    synchronized DfsReferral getDfsReferral( NtlmPasswordAuthentication auth,
+    DfsReferral getDfsReferral( NtlmPasswordAuthentication auth,
                 String path ) throws SmbException {
         String subpath, node, host;
         DfsReferral dr = new DfsReferral();
@@ -601,21 +642,23 @@ public class SmbTransport extends Transport implements SmbConstants {
                          * with challenge from 'server' to access DFS vol. */
         return dr;
     }
-    synchronized DfsReferral lookupReferral( String unc ) {
-        DfsReferral dr;
-        ListIterator iter = referrals.listIterator();
-        int i, len;
+    DfsReferral lookupReferral( String unc ) {
+        synchronized( referrals ) {
+            DfsReferral dr;
+            ListIterator iter = referrals.listIterator();
+            int i, len;
 
-        while( iter.hasNext() ) {
-            dr = (DfsReferral)iter.next();
-            len = dr.path.length();
-            for( i = 0; i < len && i < unc.length(); i++ ) {
-                if( dr.path.charAt( i ) != unc.charAt( i )) {
-                    break;
+            while( iter.hasNext() ) {
+                dr = (DfsReferral)iter.next();
+                len = dr.path.length();
+                for( i = 0; i < len && i < unc.length(); i++ ) {
+                    if( dr.path.charAt( i ) != unc.charAt( i )) {
+                        break;
+                    }
+                }
+                if( i == len && (len == unc.length() || unc.charAt( len ) == '\\')) {
+                    return dr;
                 }
-            }
-            if( i == len && (len == unc.length() || unc.charAt( len ) == '\\')) {
-                return dr;
             }
         }
 
